import base64
import hashlib
import hmac
import json
from datetime import datetime


def verify_webhook(secret: str, signature: str, payload: str) -> bool:
    if (
        not secret
        or not isinstance(secret, str)
        or not signature
        or not isinstance(signature, str)
    ):
        return False

    if not secret.startswith("atma_whsec_"):
        raise ValueError("Secret must start with 'atma_whsec_'")

    signatures = signature.split(" ")
    generated_signature = generate_webhook_signature(secret, payload)

    # Use constant-time comparison to prevent timing attacks
    return any(hmac.compare_digest(generated_signature, sig) for sig in signatures)


def generate_webhook_signature(secret: str, payload: str) -> str:
    try:
        msg = json.loads(payload)
    except json.JSONDecodeError:
        raise ValueError("Invalid payload format")

    if not isinstance(msg, dict):
        raise ValueError("Invalid payload format")

    id = msg.get("id")
    timestamp = msg.get("timestamp")

    if not id or not timestamp:
        raise ValueError("Payload must contain both 'id' and 'timestamp' fields")

    # Convert timestamp from ISO to epoch time in seconds
    dt = datetime.fromisoformat(timestamp.replace("Z", "+00:00"))
    timestamp = int(dt.timestamp())

    sig = base64.b64encode(
        hmac.new(
            secret[11:].encode("utf-8"),
            f"{id}.{timestamp}.{payload}".encode("utf-8"),
            hashlib.sha256,
        ).digest()
    ).decode("utf-8")

    return f"v1,{sig}"



from __future__ import annotations

import json
import subprocess
import tarfile
from asyncio import to_thread
from os import makedirs, remove
from os.path import join
from pathlib import Path
from shutil import rmtree
from typing import NotRequired, TypedDict, cast

from .._resource import AsyncAPIResource, SyncAPIResource
from .._types import RequestOptions

__all__ = [
    "CodeResource",
    "AsyncCodeResource",
    "CodeFolder",
    "CodeCleanupParams",
    "CodeDownloadParams",
    "CodeProposeParams",
]


# TODO: Use programmatic git instead of git command
def get_diff(path: str, base_commit: str | None = None) -> str:
    """Get the diff for the given git repository path."""
    args = ["git", "diff"]

    if base_commit:
        args.append(base_commit)

    result = subprocess.run(
        args,
        cwd=path,
        capture_output=True,
        text=True,
        check=True,
    )

    return result.stdout


class CodeFolder:
    def __init__(self, path: str | Path):
        self.path = path

    def add(self, paths: str | list[str]) -> None:
        """Add new files to git repository"""
        paths = paths if isinstance(paths, list) else [paths]
        paths = [join(self.path, p) for p in paths if p]

        subprocess.run(
            ["git", "add", "-N", *paths],
            cwd=self.path,
            check=True,
        )

    def add_all(self) -> None:
        """Add all files to git repository"""
        subprocess.run(
            ["git", "add", "-N", "."],
            cwd=self.path,
            check=True,
        )


class BaseCodeResource:
    def _path(self, task: Task) -> str:
        return f"/tmp/automa/tasks/{task['id']}"

    def _read_token(self, folder: str) -> str | None:
        token = None

        try:
            with open(
                f"{folder}/.git/automa_proposal_token", "r", encoding="utf8"
            ) as f:
                token = f.read().strip()
        except FileNotFoundError:
            pass

        return token

    def _read_base_commit(self, folder: str, type: str) -> str | None:
        base_commit = None

        try:
            with open(
                f"{folder}/.git/automa_{type}_base_commit", "r", encoding="utf8"
            ) as f:
                base_commit = f.read().strip()
        except FileNotFoundError:
            pass

        return base_commit

    def _clone_code(self, folder: str, url: str) -> None:
        path = Path(folder)

        rmtree(folder, ignore_errors=True)
        path.mkdir(parents=True, exist_ok=True)

        subprocess.run(
            ["git", "clone", "--depth=1", url, "."],
            cwd=path,
            check=True,
        )

        # Note down the base commit to send in proposal
        self._write_base_commit(folder, "proposal")

    def _extract_download(self, folder: str) -> None:
        rmtree(folder, ignore_errors=True)
        Path(folder).mkdir(parents=True, exist_ok=True)

        with tarfile.open(f"{folder}.tar.gz", "r:gz") as tar:
            tar.extractall(path=folder)

        # Note down the base commit to be used in generating diff
        self._write_base_commit(folder, "diff")

    def _write_token(self, folder: str, token: str) -> None:
        # Save the proposal token for later use
        with open(f"{folder}/.git/automa_proposal_token", "w", encoding="utf8") as f:
            f.write(token)

    def _write_base_commit(self, folder: str, type: str) -> None:
        result = subprocess.run(
            ["git", "rev-parse", "HEAD"],
            cwd=folder,
            capture_output=True,
            text=True,
            check=True,
        )

        with open(
            f"{folder}/.git/automa_{type}_base_commit", "w", encoding="utf8"
        ) as f:
            f.write(result.stdout.strip())


class CodeResource(SyncAPIResource, BaseCodeResource):
    def cleanup(self, body: CodeCleanupParams) -> None:
        folder = self._path(body["task"])

        rmtree(folder, ignore_errors=True)

        try:
            remove(f"{folder}.tar.gz")
        except FileNotFoundError:
            pass

    def download(
        self,
        body: CodeDownloadParams,
        *,
        options: RequestOptions = cast(RequestOptions, {}),
    ) -> CodeFolder:
        token = None
        path = self._path(body["task"])
        archive_path = f"{path}.tar.gz"

        with self._client.stream(
            "post",
            "/bot/code/download",
            options={
                **options,
                "json": body,
                "headers": {
                    **options.get("headers", {}),
                },
            },
        ) as response:
            token = response.headers.get("x-automa-proposal-token")
            content_type = response.headers.get("Content-Type", "")

            if content_type.startswith("application/json"):
                content = b"".join(response.iter_bytes())

                json_data = json.loads(content.decode("utf-8"))

                self._clone_code(path, json_data["url"])
            elif content_type.startswith("application/gzip"):
                rmtree(path, ignore_errors=True)
                makedirs(path, exist_ok=True)

                with open(archive_path, "wb") as archive:
                    for chunk in response.iter_bytes(chunk_size=8192):
                        archive.write(chunk)

                self._extract_download(path)
            else:
                raise ValueError(
                    f"Unexpected Content-Type: {content_type} while downloading code."
                )

        # Save the proposal token for later use
        self._write_token(path, token)

        return CodeFolder(path)

    def propose(
        self,
        body: CodeProposeParams,
        *,
        options: RequestOptions = cast(RequestOptions, {}),
    ):
        path = self._path(body["task"])
        token = self._read_token(path)
        base_commit = self._read_base_commit(path, "proposal")
        diff_base_commit = self._read_base_commit(path, "diff")

        if not token:
            raise ValueError("Failed to read the stored proposal token")

        diff = get_diff(path, base_commit or diff_base_commit)

        return self._client.post(
            "/bot/code/propose",
            body={
                **body,
                "proposal": {
                    **body.get("proposal", {}),
                    "token": token,
                    "diff": diff,
                    **({"base_commit": base_commit} if base_commit else {}),
                },
            },
            options=options,
        )


class AsyncCodeResource(AsyncAPIResource, BaseCodeResource):
    async def cleanup(self, body: CodeCleanupParams) -> None:
        folder = self._path(body["task"])

        await to_thread(rmtree, folder, ignore_errors=True)

        try:
            await to_thread(remove, f"{folder}.tar.gz")
        except FileNotFoundError:
            pass

    async def download(
        self,
        body: CodeDownloadParams,
        *,
        options: RequestOptions = cast(RequestOptions, {}),
    ) -> CodeFolder:
        token = None
        path = self._path(body["task"])
        archive_path = f"{path}.tar.gz"

        async with self._client.stream(
            "post",
            "/bot/code/download",
            options={
                **options,
                "json": body,
                "headers": {
                    **options.get("headers", {}),
                },
            },
        ) as response:
            token = response.headers.get("x-automa-proposal-token")
            content_type = response.headers.get("Content-Type", "")

            if content_type.startswith("application/json"):
                content = b""

                async for chunk in response.aiter_bytes():
                    content += chunk

                json_data = json.loads(content.decode("utf-8"))

                await to_thread(self._clone_code, path, json_data["url"])
            elif content_type.startswith("application/gzip"):
                await to_thread(rmtree, path, ignore_errors=True)
                await to_thread(makedirs, path, exist_ok=True)

                with open(archive_path, "wb") as archive:
                    async for chunk in response.aiter_bytes(chunk_size=8192):
                        await to_thread(archive.write, chunk)

                await to_thread(self._extract_download, path)
            else:
                raise ValueError(
                    f"Unexpected Content-Type: {content_type} while downloading code."
                )

        # Save the proposal token for later use
        await to_thread(self._write_token, path, token)

        return CodeFolder(path)

    async def propose(
        self,
        body: CodeProposeParams,
        *,
        options: RequestOptions = cast(RequestOptions, {}),
    ):
        path = self._path(body["task"])
        token = await to_thread(self._read_token, path)
        base_commit = await to_thread(self._read_base_commit, path, "proposal")
        diff_base_commit = await to_thread(self._read_base_commit, path, "diff")

        if not token:
            raise ValueError("Failed to read the stored proposal token")

        diff = await to_thread(get_diff, path, base_commit or diff_base_commit)

        return await self._client.post(
            "/bot/code/propose",
            body={
                **body,
                "proposal": {
                    **body.get("proposal", {}),
                    "token": token,
                    "diff": diff,
                    **({"base_commit": base_commit} if base_commit else {}),
                },
            },
            options=options,
        )


class Task(TypedDict):
    id: int


class TaskWithToken(Task):
    token: str


class CodeCleanupParams(TypedDict):
    task: Task


class CodeDownloadParams(TypedDict):
    task: TaskWithToken


class CodeProposeParams(CodeDownloadParams):
    proposal: NotRequired[Proposal]
    metadata: NotRequired[Metadata]

    class Proposal(TypedDict):
        title: NotRequired[str]
        body: NotRequired[str]

    # TODO: Add `extra_items=Any` (py 3.15)
    class Metadata(TypedDict):
        cost_in_cents: NotRequired[int]



from ._client import AsyncAutoma, Automa
from ._types import (
    GenericTaskItem,
    Org,
    ProposalTaskItem,
    Repo,
    Task,
    TaskForCode,
    TaskItem,
    WebhookEventType,
    WebhookPayload,
    WebhookProposalAcceptedData,
    WebhookProposalAcceptedPayload,
    WebhookProposalRejectedData,
    WebhookProposalRejectedPayload,
    WebhookTaskCreatedData,
    WebhookTaskCreatedPayload,
)
from .resources import (
    AsyncCodeResource,
    CodeCleanupParams,
    CodeDownloadParams,
    CodeFolder,
    CodeProposeParams,
    CodeResource,
)

__all__ = [
    "Automa",
    "AsyncAutoma",
    "AsyncCodeResource",
    "CodeResource",
    "CodeFolder",
    "CodeCleanupParams",
    "CodeDownloadParams",
    "CodeProposeParams",
    "ProposalTaskItem",
    "GenericTaskItem",
    "TaskItem",
    "Task",
    "TaskForCode",
    "Repo",
    "Org",
    "WebhookEventType",
    "WebhookTaskCreatedData",
    "WebhookTaskCreatedPayload",
    "WebhookProposalAcceptedData",
    "WebhookProposalAcceptedPayload",
    "WebhookProposalRejectedData",
    "WebhookProposalRejectedPayload",
    "WebhookPayload",
]

# Update the __module__ attribute for exported symbols so that
# error messages point to this module instead of the module
# it was originally defined in, e.g.
# automa._exceptions.NotFoundError -> automa.NotFoundError
__locals = locals()

for __name in __all__:
    if not __name.startswith("__"):
        try:
            __locals[__name].__module__ = "automa"
        except (TypeError, AttributeError):
            # Some of our exported symbols are builtins which we can't set attributes for.
            pass



from __future__ import annotations

import os

from httpx import URL
from httpx._types import HeaderTypes
from typing_extensions import override

from ._base_client import (
    AsyncAPIClient,
    SyncAPIClient,
)
from .resources import code

__all__ = [
    "Automa",
    "AsyncAutoma",
]


class Automa(SyncAPIClient):
    _default_headers: HeaderTypes

    code: code.CodeResource

    def __init__(
        self,
        *,
        base_url: str | URL | None = None,
        default_headers: HeaderTypes | None = None,
    ) -> None:
        """Construct a new synchronous Automa client instance."""
        if base_url is None:
            base_url = os.environ.get("AUTOMA_BASE_URL")
        if base_url is None:
            base_url = "https://api.automa.app"

        super().__init__(
            base_url=base_url,
        )

        self._default_headers = default_headers or {}

        self.code = code.CodeResource(self)

    @property
    @override
    def default_headers(self) -> dict[str, str]:
        return {
            **super().default_headers,
            **self._default_headers,
        }


class AsyncAutoma(AsyncAPIClient):
    _default_headers: HeaderTypes

    code: code.AsyncCodeResource

    def __init__(
        self,
        *,
        base_url: str | URL | None = None,
        default_headers: HeaderTypes | None = None,
    ) -> None:
        """Construct a new async AsyncAutoma client instance."""
        if base_url is None:
            base_url = os.environ.get("AUTOMA_BASE_URL")
        if base_url is None:
            base_url = "https://api.automa.app"

        super().__init__(
            base_url=base_url,
        )

        self._default_headers = default_headers or {}

        self.code = code.AsyncCodeResource(self)

    @property
    @override
    def default_headers(self) -> HeaderTypes:
        return {
            **super().default_headers,
            **self._default_headers,
        }



from .code import (
    AsyncCodeResource,
    CodeCleanupParams,
    CodeDownloadParams,
    CodeFolder,
    CodeProposeParams,
    CodeResource,
)

__all__ = [
    "AsyncCodeResource",
    "CodeResource",
    "CodeFolder",
    "CodeCleanupParams",
    "CodeDownloadParams",
    "CodeProposeParams",
]



from __future__ import annotations

from enum import Enum
from typing import Any, Dict, Literal, NotRequired, TypedDict, Union

from httpx._types import HeaderTypes, QueryParamTypes, RequestExtensions


class Omit:
    """In certain situations you need to be able to represent a case where a default value has
    to be explicitly removed and `None` is not an appropriate substitute, for example:

    ```py
    # as the default `Content-Type` header is `application/json` that will be sent
    client.post("/upload/files", files={"file": b"my raw file content"})

    # you can't explicitly override the header as it has to be dynamically generated
    # to look something like: 'multipart/form-data; boundary=0d8382fcf5f8c3be01ca2e11002d2983'
    client.post(..., headers={"Content-Type": "multipart/form-data"})

    # instead you can remove the default `application/json` header by passing Omit
    client.post(..., headers={"Content-Type": Omit()})
    ```
    """

    def __bool__(self) -> Literal[False]:
        return False


class RequestOptions(TypedDict, total=False):
    json: Any | None
    headers: HeaderTypes | None
    params: QueryParamTypes | None
    extensions: RequestExtensions | None


class ProposalTaskItem(TypedDict):
    id: int
    type: Literal["proposal"]
    data: ProposalTaskItemData
    bot_id: int
    repo_id: int

    class ProposalTaskItemData(TypedDict):
        prId: int
        prNumber: int
        prTitle: str
        prHead: str
        prBase: str
        prState: Literal["open", "closed"]
        prMerged: bool


class GenericTaskItem(TypedDict):
    id: int
    type: Literal["origin", "message", "repo", "bot", "activity"]
    data: Dict[str, Any]
    bot_id: NotRequired[int]
    repo_id: NotRequired[int]


TaskItem = Union[ProposalTaskItem, GenericTaskItem]


class Task(TypedDict):
    id: int
    title: str


class TaskForCode(Task):
    token: str
    items: list[TaskItem]


class Repo(TypedDict):
    id: int
    name: str
    is_private: bool


class Org(TypedDict):
    id: int
    name: str
    provider_type: Literal["github", "gitlab"]


class WebhookEventType(Enum):
    TASK_CREATED = "task.created"
    PROPOSAL_ACCEPTED = "proposal.accepted"
    PROPOSAL_REJECTED = "proposal.rejected"


class WebhookTaskCreatedData(TypedDict):
    task: TaskForCode
    repo: Repo
    org: Org


class WebhookProposalClosedData(TypedDict):
    proposal: ProposalTaskItem
    task: Task
    org: Org


WebhookProposalAcceptedData = WebhookProposalClosedData
WebhookProposalRejectedData = WebhookProposalClosedData


class WebhookTaskCreatedPayload(TypedDict):
    id: str
    timestamp: str
    type: Literal[WebhookEventType.TASK_CREATED]
    data: WebhookTaskCreatedData


class WebhookProposalAcceptedPayload(TypedDict):
    id: str
    timestamp: str
    type: Literal[WebhookEventType.PROPOSAL_ACCEPTED]
    data: WebhookProposalAcceptedData


class WebhookProposalRejectedPayload(TypedDict):
    id: str
    timestamp: str
    type: Literal[WebhookEventType.PROPOSAL_REJECTED]
    data: WebhookProposalRejectedData


WebhookPayload = Union[
    WebhookTaskCreatedPayload,
    WebhookProposalAcceptedPayload,
    WebhookProposalRejectedPayload,
]



from __future__ import annotations

import time
from typing import TYPE_CHECKING

import anyio

if TYPE_CHECKING:
    from ._client import AsyncAutoma, Automa


class SyncAPIResource:
    _client: Automa

    def __init__(self, client: Automa) -> None:
        self._client = client

    def _sleep(self, seconds: float) -> None:
        time.sleep(seconds)


class AsyncAPIResource:
    _client: AsyncAutoma

    def __init__(self, client: AsyncAutoma) -> None:
        self._client = client

    async def _sleep(self, seconds: float) -> None:
        await anyio.sleep(seconds)



from src.automa.bot.webhook import generate_webhook_signature, verify_webhook


def test_returns_false_if_secret_is_not_a_string():
    result = verify_webhook(1, "signature", "{}")  # ty:ignore[invalid-argument-type]

    assert result is False


def test_returns_false_if_secret_is_empty():
    result = verify_webhook("", "signature", "{}")

    assert result is False


def test_returns_false_if_signature_is_not_a_string():
    result = verify_webhook("secret", 1, "{}")  # ty:ignore[invalid-argument-type]

    assert result is False


def test_returns_false_if_signature_is_empty():
    result = verify_webhook("secret", "", "{}")

    assert result is False


def test_throws_error_if_secret_does_not_start_with_atma_whsec():
    try:
        verify_webhook("invalid_secret", "signature", "{}")
    except ValueError as e:
        assert str(e) == "Secret must start with 'atma_whsec_'"


def test_throws_error_if_payload_is_not_a_valid_json():
    try:
        verify_webhook("atma_whsec_secret", "signature", "not a json")
    except ValueError as e:
        assert str(e) == "Invalid payload format"


def test_throws_error_if_payload_is_not_a_dict():
    try:
        verify_webhook("atma_whsec_secret", "signature", '["not", "a", "dict"]')
    except ValueError as e:
        assert str(e) == "Invalid payload format"


def test_throws_error_if_payload_does_not_contain_id():
    try:
        verify_webhook(
            "atma_whsec_secret", "signature", '{"timestamp": "2023-10-01T00:00:00Z"}'
        )
    except ValueError as e:
        assert str(e) == "Payload must contain both 'id' and 'timestamp' fields"


def test_throws_error_if_payload_does_not_contain_timestamp():
    try:
        verify_webhook("atma_whsec_secret", "signature", '{"id": "1"}')
    except ValueError as e:
        assert str(e) == "Payload must contain both 'id' and 'timestamp' fields"


def test_returns_false_if_signature_is_wrong():
    result = verify_webhook(
        "atma_whsec_secret",
        "signature",
        '{"id": "1", "timestamp": "2023-10-01T00:00:00Z"}',
    )

    assert result is False


def test_returns_true_if_signature_is_correct():
    result = verify_webhook(
        "atma_whsec_secret",
        "v1,/VB+gt7CALYoWLO0Pym4MfxP3l4a29h1OYQcfJ59D+E=",
        '{"id": "1", "timestamp": "2023-10-01T00:00:00Z"}',
    )

    assert result is True


def test_returns_true_if_second_signature_is_correct():
    result = verify_webhook(
        "atma_whsec_secret",
        "bad_signature v1,/VB+gt7CALYoWLO0Pym4MfxP3l4a29h1OYQcfJ59D+E=",
        '{"id": "1", "timestamp": "2023-10-01T00:00:00Z"}',
    )

    assert result is True


def test_verifies_the_generated_signature():
    # Generate signature
    signature = generate_webhook_signature(
        "atma_whsec_secret", '{"id": "1", "timestamp": "2023-10-01T00:00:00Z"}'
    )

    # Verify with the same parameters
    result = verify_webhook(
        "atma_whsec_secret",
        signature,
        '{"id": "1", "timestamp": "2023-10-01T00:00:00Z"}',
    )

    assert result is True



from __future__ import annotations

import asyncio
from contextlib import asynccontextmanager, contextmanager
from types import TracebackType
from typing import (
    Any,
    AsyncIterator,
    Generic,
    Iterator,
    TypeVar,
    Union,
    cast,
)

from httpx import URL, AsyncClient, Client, Response
from httpx._types import HeaderTypes

from ._types import RequestOptions

_T = TypeVar("_T")

_HttpxClientT = TypeVar("_HttpxClientT", bound=Union[Client, AsyncClient])


class BaseClient(Generic[_HttpxClientT]):
    _client: _HttpxClientT
    _base_url: URL

    def __init__(
        self,
        *,
        base_url: str | URL,
    ) -> None:
        self.base_url = base_url

    def _enforce_trailing_slash(self, url: URL) -> URL:
        if url.raw_path.endswith(b"/"):
            return url
        return url.copy_with(raw_path=url.raw_path + b"/")

    @property
    def base_url(self) -> URL:
        return self._base_url

    @base_url.setter
    def base_url(self, url: str | URL) -> None:
        self._base_url = self._enforce_trailing_slash(
            url if isinstance(url, URL) else URL(url)
        )

    @property
    def default_headers(self) -> HeaderTypes:
        return {
            "Accept": "application/json",
            "Content-Type": "application/json",
        }


class SyncHttpxClientWrapper(Client):
    def __del__(self) -> None:
        if self.is_closed:
            return

        try:
            self.close()
        except Exception:
            pass


class SyncAPIClient(BaseClient[Client]):
    _client: Client

    def __init__(
        self,
        *,
        base_url: str | URL,
    ) -> None:
        super().__init__(
            base_url=base_url,
        )
        self._client = SyncHttpxClientWrapper(
            base_url=base_url,
        )

    def is_closed(self) -> bool:
        return self._client.is_closed

    def close(self) -> None:
        """Close the underlying HTTPX client.

        The client will *not* be usable after this.
        """
        # If an error is thrown while constructing a client, self._client
        # may not be present
        if hasattr(self, "_client"):
            self._client.close()

    def __enter__(self: _T) -> _T:
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        self.close()

    def get(
        self,
        path: str,
        *,
        options: RequestOptions = cast(RequestOptions, {}),
    ) -> Response:
        return self.method("get", path, options=options)

    def post(
        self,
        path: str,
        *,
        body: Any | None = None,
        options: RequestOptions = cast(RequestOptions, {}),
    ) -> Response:
        return self.method(
            "post",
            path,
            options={
                **options,
                "json": body,
            },
        )

    def method(
        self,
        method: str,
        path: str,
        *,
        options: RequestOptions = cast(RequestOptions, {}),
    ) -> Response:
        options["headers"] = {**self.default_headers, **options.get("headers", {})}

        response = self._client.request(method, path, **options)

        if response.is_error:
            data = response.json()
            raise Exception(data.get("message") or data.get("error"))

        return response

    @contextmanager
    def stream(
        self,
        method: str,
        path: str,
        *,
        options: RequestOptions = cast(RequestOptions, {}),
    ) -> Iterator[Response]:
        options["headers"] = {**self.default_headers, **options.get("headers", {})}

        with self._client.stream(method, path, **options) as response:
            if response.is_error:
                response.read()
                data = response.json()
                raise Exception(data.get("message") or data.get("error"))

            yield response


class AsyncHttpxClientWrapper(AsyncClient):
    def __del__(self) -> None:
        if self.is_closed:
            return

        try:
            asyncio.get_running_loop().create_task(self.aclose())
        except Exception:
            pass


class AsyncAPIClient(BaseClient[AsyncClient]):
    _client: AsyncClient

    def __init__(
        self,
        *,
        base_url: str | URL,
    ) -> None:
        super().__init__(
            base_url=base_url,
        )
        self._client = AsyncHttpxClientWrapper(
            base_url=base_url,
        )

    def is_closed(self) -> bool:
        return self._client.is_closed

    async def close(self) -> None:
        """Close the underlying HTTPX client.

        The client will *not* be usable after this.
        """
        await self._client.aclose()

    async def __aenter__(self: _T) -> _T:
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        await self.close()

    async def get(
        self,
        path: str,
        *,
        options: RequestOptions = cast(RequestOptions, {}),
    ) -> Response:
        return await self.method("get", path, options=options)

    async def post(
        self,
        path: str,
        *,
        body: Any | None = None,
        options: RequestOptions = cast(RequestOptions, {}),
    ) -> Response:
        return await self.method(
            "post",
            path,
            options={
                **options,
                "json": body,
            },
        )

    async def method(
        self,
        method: str,
        path: str,
        *,
        options: RequestOptions = cast(RequestOptions, {}),
    ) -> Response:
        options["headers"] = {**self.default_headers, **options.get("headers", {})}

        response = await self._client.request(method, path, **options)

        if response.is_error:
            data = response.json()
            raise Exception(data.get("message") or data.get("error"))

        return response

    @asynccontextmanager
    async def stream(
        self,
        method: str,
        path: str,
        *,
        options: RequestOptions = cast(RequestOptions, {}),
    ) -> AsyncIterator[Response]:
        options["headers"] = {**self.default_headers, **options.get("headers", {})}

        async with self._client.stream(method, path, **options) as response:
            if response.is_error:
                await response.aread()
                data = response.json()
                raise Exception(data.get("message") or data.get("error"))

            yield response
